// Copyright (c) 2016-2024 CloudCrane. All rights reserved.
//
// This work is licensed under the MIT License.
// See the LICENSE file in the project root for more information.

package fullstack

import (
	"testing"
	"time"

	_ "github.com/mattn/go-sqlite3"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
	"xbus.io/go-xbus/v4"
	"xbus.io/go-xbus/v4/api"
	"xbus.io/go-xbus/v4/api/control"

	"xbus.io/xbus/testutils"
)

func TestRegistration(t *testing.T) {
	ft := InitFullstack(t, true)
	defer ft.Shutdown()

	t.Run("Invalid account name", func(t *testing.T) {
		options := xbus.COptions{
			Name:        "valide",
			ConnOptions: ft.Server.NATSConnectOptions(),
		}
		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		assert.NoError(t, client.Connect())

		var req api.RegistrationRequest
		_, err := client.API.SvcRegistration.Register(&req)
		assert.Equal(t, "CLIENT error: Invalid account name: ''", err.Error())

		req.Name = "nom invalide"
		_, err = client.API.SvcRegistration.Register(&req)
		assert.Equal(t, "CLIENT error: Invalid account name: 'nom invalide'", err.Error())

		req.Name = "valide"
		_, err = client.API.SvcRegistration.Register(&req)
		assert.Nil(t, err)
	})

	t.Run("Actor configuration preserved", func(t *testing.T) {
		options := xbus.COptions{
			Name:        "valide",
			ConnOptions: ft.Server.NATSConnectOptions(),
			ActorList: []xbus.ActorOptions{{
				Actor: &api.Actor{
					Name:  "toto",
					Kind:  api.Actor_CONSUMER,
					Roles: []string{"toto"},
				},
			}},
		}

		logger := testutils.GetTestLogger(t)
		client := xbus.NewClient(options, logger.Logger, logger.LoggerOutput)

		assert.NoError(t, client.Connect())

		if err := client.CreateKey(true); err != nil {
			require.NotErrorIs(t, err, xbus.ErrAlreadyExists)
		}
		require.NoError(t, client.CreateCSR("localhost", "xbus", "xbusctl", true))

		_, err := client.Register()
		assert.Nil(t, err)

		client.Close()

		account, err := ft.Client.AccountFind("valide")
		require.NoError(t, err)

		_, err = ft.Client.AccountAccept(account.GetIDAsUUID(), time.Now().Add(10*time.Second))
		require.NoError(t, err)

		actor, err := ft.Client.ActorFind("toto")
		require.NoError(t, err)
		require.Empty(t, actor.Config)

		_, err = ft.Client.ActorAccept(actor.GetIDAsUUID())
		require.NoError(t, err)

		actor.Config = append(actor.Config, &api.ConfigEntry{
			Key:   "key",
			Value: "value",
		})

		_, err = ft.Client.Control.Actor.Update(&control.ActorRequest{
			ActorList: []*api.Actor{{
				ID:     actor.ID,
				Config: actor.Config,
				Status: api.Actor_ENABLED,
			}},
		})
		require.NoError(t, err)

		_, err = ft.Client.ActorAccept(actor.GetIDAsUUID())
		require.NoError(t, err)

		actor, err = ft.Client.ActorFind("toto")
		require.NoError(t, err)
		require.NotEmpty(t, actor.Config)
	})
}
